class ListExtension(list):
    def __new__(cls, other=None):
        if isinstance(other, cls):
            return other
        return super().__new__(cls)

    def __init__(self, other=None):
        if other is None:
            other = []
        super().__init__(other)

    def find(self, lmbd, *args, **kwargs):
        for item in self:
            if lmbd(item, *args, **kwargs):
                return item

    def join(self, separator: str, prefix="", postfix=""):
        string = ""
        for iterable, item in enumerate(self):
            string = f"{string}{prefix}{item}{postfix}"
            if iterable + 1 != len(self):
                string += separator
        return string

    def __call__(self):
        return ListExtension()

    def findall(self, lmbd, *args, **kwargs):
        lst = self()
        for item in self:
            if lmbd(item, *args, **kwargs):
                lst.append(item)
        return lst

    @classmethod
    def indexList(cls, list=None):
        iterate = getattr(list, "dictionary", list) or cls
        l = cls()
        for index in range(len(iterate)):
            l.append(index)
        return l

    def all(self, function, *args, **kwargs):
        for i in self:
            if not function(i, *args, **kwargs):
                return False
        return True

    def has(self, item, returnIndex=False):
        for i, iterator in enumerate(self):
            if hasattr(iterator, "has") and callable(iterator.has):
                if iterator.has(item):
                    return True if not returnIndex else i
            if iterator == item:
                return True if not returnIndex else i
        return False if not returnIndex else -1

    def indexOf(self, item):
        return self.has(item, True)

    def first(self):
        return self[0]

    def __getitem__(self, key):
        if isinstance(key, slice):
            return super().__getitem__(key)
        if key < len(self):
            return super().__getitem__(key)
        return None

    def filter(self, lmbd):
        instance = ListExtension()
        for i in self:
            if lmbd(i):
                instance.append(i)
        return instance

    def forEach(self, lmbd):
        for item in self:
            lmbd(item)

    def includes(self, value):
        return value in self

    @classmethod
    def byList(cls, lst):
        return cls(lst)

    def append(self, value):
        super().append(value)
        return self

    def copy(self):
        return ListExtension(self)

    def map(self, lmbd, copy=False, *args, **kwargs):
        save = self if not copy else self()
        for i, _ in enumerate(self):
            if copy:
                save.append(lmbd(_, *args, **kwargs))
            else:
                save[i] = lmbd(_, *args, **kwargs)
        return save

    def __add__(self, other):
        if type(other) is list:
            self += other
        else:
            self.append(other)
        return self



# command + after func
import difflib
import inspect

from . import database
from .jsonExtension import StructByAction


class AfterFunc(database.Struct):
    save_by = "user_id"
    user_id = database.Sqlite3Property("", "not null unique")
    after_name = ""
    args = []


class AbstractAfterFunc(object):
    def __new__(cls, name, func=None):
        if (instance := after_func_poll.get(name)) is None:
            return super().__new__(cls)
        return instance

    def __init__(self, name, func=None) -> None:
        if after_func_poll.get(name) is None:
            self.name = name
            self.text_matchers = {}
            after_func_poll[name] = self
        if func is not None:
            self.text_matchers["default"] = func


command_poll = []  # MutableList<Command>
after_func_poll = {}


# data class for commands
class Command(object):
    def __init__(self, name, aliases, fixTypo, callableItem):
        self.name = name
        self.aliases = aliases
        self.fixTypo = fixTypo
        self.callable = callableItem


def wait(name, uID, function):
    after_func(name)(function)
    set_after(name, uID)


def command(name, fixTypo=True, aliases=None):
    if aliases is None:
        aliases = []

    def func_wrap(func):
        command_poll.append(Command(name, aliases, fixTypo, func))

    return func_wrap


start_command = command("начать", aliases=["start", "меню", "бот", "старт"])


def after_func_from_lambda(name, func):
    after_func(name)(func)


def after_func(name):
    def func_wrap(func):
        AbstractAfterFunc(name, func)
    return func_wrap


def after_text_matcher(name, text):
    def func_wrap(func):
        function = AbstractAfterFunc(name)
        function.text_matchers[text] = func
    return func_wrap


def set_after(name, uID, args=None):
    if args is None:
        args = []
    struct = AfterFunc(after_name=name, user_id=uID)
    struct.after_name = name
    struct.args = args


def call_command(function, *args):
    function_args = inspect.getfullargspec(function).args
    if len(function_args) == 1:
        return function(args[0])
    function(*args)


def execute_command(botClass):
    selected = database.db.select_one_struct("select * from after_func where user_id = ?",
                                             [botClass.user.id])

    if selected is not None and selected.after_name != "null":
        tmpAfterName = selected.after_name
        selected.after_name = "null"
        doNotReset = False
        if tmpAfterName in after_func_poll:
            after_func = after_func_poll[tmpAfterName]
            call = after_func.text_matchers.get(
                botClass.text) or after_func.text_matchers.get("default")
            if call is not None:
                doNotReset = call(botClass) if (isinstance(selected.args,
                                                           StructByAction) and not selected.args.dictionary) or not selected.args else \
                    call(botClass, selected.args)
                if doNotReset is None or call.__name__ == "<lambda>" or not isinstance(doNotReset, bool):
                    doNotReset = False
                if doNotReset:
                    selected.after_name = tmpAfterName
        return
    tmpCmd = []
    # loop over arguments
    for i in botClass.txtSplit:
        tmpCmd.append(i)
        name = " ".join(tmpCmd)
        # prefer names over fixed commands
        for cmd in command_poll:
            if cmd.name == name or name in cmd.aliases:
                call_command(cmd.callable, botClass, botClass.args)
                return
        for cmd in command_poll:
            if not cmd.fixTypo:
                continue
            matches = difflib.get_close_matches(name, cmd.aliases, cutoff=0.7)
            if not matches:
                continue
            call_command(cmd.callable, botClass, botClass.args)
            return



from . import (user, cmd, database, events)
import json
import vk_api
from vk_api.longpoll import VkEventType, VkLongPoll
from .listExtension import ListExtension
from .stringExtension import StringExtension
from .thread import Thread
from .database import config
import re
from . import imports
imports.ImportTools(["Structs"])


class LongPoll(VkLongPoll):
    def __init__(self, instance, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance = instance

    def listen(self):
        while True:
            try:
                self.instance.check_tasks()
                updates = self.check()
                for event in updates:
                    yield event
            except:
                pass


class AbstractChatLongPoll(Thread):
    def __init__(self, config, **kwargs) -> None:
        self.config = config
        self.vk_session = vk_api.VkApi(token=config["vk_api_key"])
        self.longpoll = LongPoll(self, self.vk_session)
        self.vk = self.vk_session.get_api()
        self.db = database.Database(self.config)
        super().__init__(**kwargs)

    def parse_attachments(self):
        for attachmentList in self.attachments_last_message:
            attachment_type = attachmentList['type']
            attachment = attachmentList[attachment_type]
            access_key = attachment.get("access_key")
            if attachment_type != "sticker":
                self.attachments.append(
                    f"{attachment_type}{attachment['owner_id']}_{attachment['id']}") if access_key is None \
                    else self.attachments.append(
                    f"{attachment_type}{attachment['owner_id']}_{attachment['id']}_{access_key}")
            else:
                self.sticker_id = attachment["sticker_id"]

    def write(self, user_id, *args, **kwargs):
        user.User(user_id, vk=self.vk).write(*args, **kwargs)

    def reply(self, *args, **kwargs):
        return self.user.write(*args, **kwargs)

    def on_message(self, event):
        pass

    def run(self):
        for event in self.longpoll.listen():
            if event.type == VkEventType.MESSAGE_NEW and event.to_me and not event.from_me and not event.from_group and not event.from_chat:
                self.attachments = ListExtension()
                self.sticker_id = None
                self.user = user.User(event.user_id, vk=self.vk)
                self.raw_text = StringExtension(event.message.strip())
                self.event = event
                if hasattr(self.event, "payload") and self.event.payload is not None:
                    self.event.payload = json.loads(self.event.payload)
                self.text = StringExtension(self.raw_text.lower().strip())
                self.txtSplit = self.text.split()
                self.command = self.txtSplit[0] if len(
                    self.txtSplit) > 0 else ""
                self.args = self.txtSplit[1:]
                self.messages = self.user.messages.getHistory(count=3)["items"]
                self.last_message = self.messages[0]
                self.attachments_last_message = self.last_message["attachments"]
                self.parse_attachments()
                self.on_message(event)


class BotLongPoll(AbstractChatLongPoll):
    def on_start(self):
        events.emit("start")
        self.started = True

    def __init__(self, c=None, **kwargs) -> None:
        super().__init__(c or config, **kwargs)
        imports.ImportTools(["packages", "Structs"])
        self.group_id = "-" + re.findall(r'\d+', self.longpoll.server)[0]

    def wait(self, x, y):
        return cmd.set_after(x, self.user.id, y)

    def set_after(self, x, y=None):
        if y is None:
            y = []
        cmd.set_after(x, self.user.id, y)

# for future uses?


class UserLongPoll(AbstractChatLongPoll):
    pass



from vk_api.exceptions import ApiError

from .methodExecutor import MethodExecutor
from .keyboard import Keyboard
import vk_api
from . import thread
from .cmd import set_after

class User(MethodExecutor):
    # method to insert keyword map
    user_id_methods = {
        "messages.getHistory": "user_id", "users.get": "user_ids"}

    def __new__(cls, user_id, fields = None, **kwargs):
        try:
            if fields is None:
                fields = "photo_id,photo_50"
            vk = kwargs.get("vk")
            instance = super(User, cls).__new__(cls)
            if vk is None:
                vk = thread.main_thread.vk
                kwargs["vk"] = vk
            instance.vk = vk
            get = vk.users.get(user_ids=user_id, fields=fields)[0]
            instance.request = get
            return instance
        except:
            return None

    def __init__(self, user_id=None, fields = None, vk=None):
        self.avatar = f"photo{self.request['photo_id']}" if self.request.get(
            'photo_id') is not None else ""
        self.user_name = f"{self.request['first_name']} {self.request['last_name']}"
        self.id = str(self.request["id"])

        super().__init__(self.vk, self.on_method_execute)

    def write(self, message, keyboard=None, after=None, after_args=None, **kwargs):
        if keyboard is not None:
            kwargs["keyboard"] = Keyboard.byKeyboard(keyboard)
        try:
            self._vk.messages.send(user_id=self.id, message=message, random_id=vk_api.utils.get_random_id(),
                                   **kwargs)
        except ApiError[50] as e:
            print(e)
        finally:
            if after is not None:
                set_after(after, self.id, after_args)

    def on_method_execute(self, method, kwargs):
        if method in User.user_id_methods:
            kwargs[User.user_id_methods[method]] = self.id

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, type(self)):
            raise NotImplementedError
        return self.id == o.id

    def __hash__(self) -> int:
        return int(self.id)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id})"

    @property
    def mention(self):
        return f"@id{self.id}({self.user_name})"



import os
import re
import sqlite3
from collections import namedtuple
from typing import AnyStr, Any, Iterable
from sqlite3 import Row

from . import jsonExtension, thread, timeExtension
from .listExtension import ListExtension


def getter(x: Any, attr: AnyStr): return getattr(x, attr, x)


DEFAULT_CONFIG = """
{
    "db_file": "data/db.sqlite3",
    "db_backups": false,
    "db_backups_folder": "backups/",
    "db_backup_interval": 43200,
    "sync_timezone": "Europe/Moscow",
    "vk_api_key": ""
}
"""

config = jsonExtension.loadAdvanced("config.json", content=DEFAULT_CONFIG)


def attrgetter(x: Any): return getter(x, "value")


def formAndExpr(baseSql, argsList, getattrFrom, add):
    for i, k in enumerate(add):
        baseSql += f"{k}=?"
        argsList.append(getattr(getattrFrom, k))
        if i != len(add) - 1:
            baseSql += " and "
    return baseSql, argsList


def to_sneak_case(string):
    return re.sub(r'(?<!^)(?=[A-Z])', '_', string).lower()


# Handle all stuff behind Struct instances
def convert_to_list_if_needed(element):
    if not isinstance(element, list):
        return ListExtension([element])
    else:
        return ListExtension(element)


class Struct(object):
    table_map = {}

    @classmethod
    def extract_table_name(cls):
        if hasattr(cls, "table_name"):
            return getattr(cls, "table_name")
        else:
            return to_sneak_case(cls.__name__)

    @classmethod
    def extract_save_by(cls):
        if hasattr(cls, "save_by"):
            return convert_to_list_if_needed(attrgetter(cls.save_by))
        for k, v in vars(cls).items():
            if isinstance(v, Sqlite3Property) and "unique" in v.type:
                return ListExtension(k)

    def __init_subclass__(cls) -> None:
        cls.table_name = cls.extract_table_name()
        cls.save_by = cls.extract_save_by()
        cls.table_map[cls.table_name] = cls
        return super().__init_subclass__()

    def define_db(self):
        if hasattr(self, "use_db"):
            self.db = Database.db_cache[self.use_db]
        else:
            self.db = db

    def __new__(cls, **kwargs):
        create_new = kwargs.get("create_new", True)
        cls.define_db(cls)
        instance = super().__new__(cls)
        instance.setattr("old_struct", None, ignore_duplicates=False)
        if kwargs:
            expr = f"select * from {cls.table_name} where "
            args = []
            n = namedtuple('Struct', kwargs.keys())(**kwargs)
            expr, args = formAndExpr(expr, args, n, cls.save_by)
            old_struct = cls.db.select_one_struct(expr, args)
            if old_struct is None and create_new is False:
                return None
            instance.old_struct = old_struct
        return instance

    def __init__(self, create_new=None, **kwargs) -> None:
        if kwargs:
            for key, value in kwargs.items():
                setattr(self, key, value)
            if self.save_by.all(lambda it: it in kwargs.keys()):
                if self.old_struct is None:
                    keys, values = ListExtension(), ListExtension()
                    vrs = self.vars()
                    for key, value in vrs.items():
                        keys.append(key)
                        values.append(kwargs[key]) if kwargs.get(
                            key) is not None else values.append(value)
                    insert_string = f"insert or ignore into {self.table_name} ({','.join(keys)}) values ({values.map(lambda _: '?', copy=True).join(',')})"
                    self.db.execute(insert_string, values)
                    variables = self.vars()
                    self.fill(variables.keys(), variables)
                else:
                    keys = kwargs.keys()
                    values = ListExtension(kwargs.values()).filter(
                        lambda it: it not in self.save_by)
                    d = dict(zip(keys, values))
                    for k, v in d.items():
                        self.old_struct.setattr(k, v)
                    variables = self.old_struct.vars()
                    self.fill(variables.keys(), variables)
                self.initialized = True

        super().__init__()

    def boundStructByAction(self, key, data):
        data = getattr(data, "dictionary", data)
        structByAction = jsonExtension.StructByAction(data)
        structByAction.action = lambda _: self.db.save_struct_by_action(
            self.table_name, key, structByAction, self.save_by, self)
        return structByAction

    def destroy(self):
        lst = []
        sql = f"delete from {self.table_name} where "
        sql, lst = formAndExpr(sql, lst, self, self.save_by)
        self.db.execute(sql, lst)

    def setattr(self, key, value, write_to_database=True, ignore_duplicates=True):
        prev = getattr(self, key, None)
        if prev == value and ignore_duplicates:
            return
        if write_to_database and getattr(self, "initialized", False):
            if isinstance(prev, jsonExtension.StructByAction):
                super().__setattr__(key, self.boundStructByAction(key, value))
                getattr(self, key).action(None)
            else:
                self.db.write_struct(self, key, value)
                super().__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    def vars(cls):
        attrs = {k: getattr(cls, k) for k in dir(cls)}
        return {k: v for k, v in attrs.items() if not k.startswith(
                "__") and k not in ["table_name", "save_by", "initialized", "table_map", "use_db", "db", "old_struct"] and not callable(v)}

    def fill(self, keys, getitemfrom):
        for k in keys:
            v = getitemfrom[k]
            attr = v
            data, value = jsonExtension.isDeserializable(v)
            if isinstance(v, list) or isinstance(v, dict):
                attr = self.boundStructByAction(k, v)
            elif value:
                attr = self.boundStructByAction(k, data)
            if isinstance(getattr(self, k), bool):
                attr = self.str2bool(attr)
            self.setattr(k, attr, False)

    def __setattr__(self, name: str, value: Any) -> None:
        self.setattr(name, value)

    def str2bool(self, v):
        if hasattr(v, "lower"):
            return v.lower() in ("true", "1")
        else:
            return v


class Sqlite3Property(object):
    def __init__(self, x: Any, y: AnyStr):
        self.value = x
        self.type = y


class Database(object):

    typeToTypeCache = {str: "text", dict: "str", list: "str",
                       float: "real", type(None): "null", int: "int", bool: "bool"}
    db_cache = {}

    def __new__(cls, settings, **kwargs):
        file = settings["db_file"]
        if (instance := cls.db_cache.get(file)) is None:
            instance = super().__new__(cls)
            cls.db_cache[os.path.basename(
                os.path.splitext(file)[0])] = instance  # short path
            cls.db_cache[file] = instance
            return instance
        return instance

    def __init__(self, settings: dict, **kwargs):
        self.settings = settings
        self.backup_folder = self.settings["db_backups_folder"]
        folder = os.path.split(self.settings["db_file"])[0]

        if not os.path.exists(folder):
            os.makedirs(folder)
        self.file = self.settings["db_file"]
        self.db = sqlite3.connect(
            self.settings["db_file"], check_same_thread=False, **kwargs)
        self.row_factory = sqlite3.Row
        self.db.row_factory = self.row_factory
        self.cursor = self.db.cursor()

        is_main_table = self.settings["db_file"] == config["db_file"]

        for struct in Struct.table_map.values():
            if not is_main_table and not hasattr(struct, "use_db") or hasattr(struct, "use_db") and self.db_cache[struct.use_db] != self:
                return
            iterable = -1
            rows = []
            variables = struct.vars(struct)
            for key, value in variables.items():
                iterable += 1
                real_value = attrgetter(value)
                rows.append(
                    f"{key} {self.convert_type(real_value)} {getattr(value, 'type', '')} default \"{real_value}\"")
            self.execute(
                f"create table if not exists {struct.table_name} ({', '.join(rows)})")
            table_fields = self.get_column_names(struct.table_name)
            for iterable, field in enumerate(variables.keys()):
                if field not in table_fields:
                    self.execute(
                        f"alter table {struct.table_name} add column {rows[iterable]}")
            for field in list(table_fields):
                if field not in variables:
                    self.execute(
                        f"alter table {struct.table_name} drop column {field}")
        if self.settings["db_backups"]:
            thread.every(self.settings["db_backup_interval"], name="Backup")(
                self.backup)

        global db
        if config["db_file"] == self.settings["db_file"]:
            db = self

    def backup(self):
        if not self.settings["db_backups"]:
            return

        rawName = self.file.split("/")[-1]
        manager = thread.ThreadManager()
        manager.changeInterval(
            "Backup", self.settings["db_backup_interval"])
        backup_table = sqlite3.connect(
            f"{self.backup_folder}backup_{timeExtension.now()}_{rawName}")
        self.db.backup(backup_table)

    def select(self, query: AnyStr, args=None):
        if args is None:
            args = []
        self.cursor.execute(query, args)
        return self.cursor.fetchall()

    def select_one(self, query: AnyStr, *args):
        if isinstance(args, list):
            self.cursor.execute(query, [str(x) for x in args])
        else:
            self.cursor.execute(query, *args)
        return self.cursor.fetchone()

    def write_struct(self, structToWrite: Struct, changedKey: AnyStr, newValue: Any):
        table = structToWrite.table_name
        unique_fields = Struct.table_map[table].save_by
        sql = f"update or ignore {table} set {changedKey} = ? where "
        argsList = [newValue]
        sql, argsList = formAndExpr(
            sql, argsList, structToWrite, unique_fields)
        self.execute(sql, argsList)

    def select_one_struct(self, query: AnyStr, *args: tuple or jsonExtension.StructByAction,
                          selectedStruct: Row = None,
                          fromSerialized=None, table_name=None):
        table_name = self.parse_table_name(query, table_name)
        struct = self.select_one(
            query, *args) if selectedStruct is None else selectedStruct
        if struct is None:
            return None
        if isinstance(args, jsonExtension.StructByAction):
            args = args.dictionary
        if not isinstance(table_name, str):
            raise Exception(
                f"Table name's type is not string (table_name was not provided correctly?)\n{query=}\n{args=}\n{table_name=}")
        myStruct: Struct = Struct.table_map[table_name](
        ) if fromSerialized is None else fromSerialized
        if struct is None:
            return None
        myStruct.fill(struct.keys(), struct)
        myStruct.setattr("initialized", True, write_to_database=False)
        return myStruct

    def select_all_structs(self, query: AnyStr, *args) -> ListExtension:
        structs = ListExtension.byList(self.select(query, *args))
        return ListExtension.byList([self.select_one_struct(query, *args, selectedStruct=x) for x in structs])

    def save_struct_by_action(self, table_name: AnyStr, key: Any, value: Any,
                              unique_field: Iterable, parent_struct: Struct):
        baseSql = f"update {table_name} set {key} = ? where "
        argsList = [jsonExtension.json.dumps(value.dictionary)]
        baseSql, argsList = formAndExpr(
            baseSql, argsList, parent_struct, unique_field)
        self.execute(baseSql, argsList)

    def execute(self, query: AnyStr, args=None):
        if args is None:
            args = []
        for i, k in enumerate(args):
            if type(k) is dict or type(k) is list:
                args[i] = jsonExtension.json.dumps(k)
            elif type(k) is jsonExtension.StructByAction:
                args[i] = jsonExtension.json.dumps(k.dictionary)
        self.cursor.execute(query, args)
        self.db.commit()
        return self.cursor

    def get_column_names(self, table: AnyStr):
        select = self.cursor.execute(f"select * from {table}")
        return [x[0] for x in select.description]

    def parse_table_name(self, query, fromCached=None):
        if fromCached is None:
            return list(tables_in_query(query))[0]
        return fromCached

    def get_table_names(self):
        return [x["name"] for x in self.select("SELECT name FROM sqlite_master WHERE type='table'")]

    @staticmethod
    def convert_type(value):
        return Database.typeToTypeCache[type(value)]


db: Database = None


# https://grisha.org/blog/2016/11/14/table-names-from-sql/
def tables_in_query(sql_str):
    q = re.sub(r"/\*[^*]*\*+(?:[^*/][^*]*\*+)*/", "", sql_str)
    lines = [line for line in q.splitlines(
    ) if not re.match(r"^\s*(--|#)", line)]
    q = " ".join([re.split(r"--|#", line)[0] for line in lines])
    tokens = re.split(r"[\s)(;]+", q)
    result = set()
    get_next = False
    for tok in tokens:
        if get_next:
            if tok.lower() not in ["", "select"]:
                result.add(tok)
            get_next = False
        get_next = tok.lower() in ["from", "join"]

    return result



from vk_api.exceptions import ApiError
from vk_api import VkApi


def empty(*args, **kwargs):
    pass


class MethodExecutor(object):
    def __init__(self, vk, on_method_execute, on_error=None, method=None) -> None:
        self._method = method
        self.on_method_execute = on_method_execute or empty
        self.on_error = on_error or empty
        self._vk = vk
        # super().__init__()

    def __getattr__(self, method):  # str8 up
        if '_' in method:
            m = method.split('_')
            method = m[0] + ''.join(i.title() for i in m[1:])
        return MethodExecutor(
            self._vk,
            self.on_method_execute,
            self.on_error,
            (self._method + '.' if self._method else '') + method
        )

    def __call__(self, **kwargs):
        if self._method is not None:
            if self.on_method_execute is not None:
                shouldExecute = self.on_method_execute(self._method, kwargs)
                shouldExecute = shouldExecute or True
            if not shouldExecute:
                return
            self._vk._method = self._method
            try:
                tmpReturn = self._vk.__call__(**kwargs)
                return tmpReturn
            except Exception as e:
                self.on_error(e)
            finally:
                self._vk._method = None
                self._method = None


class AuthBasedMethodExecutor(MethodExecutor):
    def __init__(self, token, invalid_callback=None) -> None:
        self._valid = True
        self.invalid_callback = invalid_callback or empty
        vk = VkApi(token=token).get_api()
        super().__init__(vk, self.on_method_execute, self.on_error)

    def on_method_execute(self, *args, **kwargs):
        return self._valid

    def on_error(self, e):
        if isinstance(e, ApiError) and e.code == 5:
            self._valid = False
            self.invalid_callback()



from .listExtension import ListExtension
import random

from vk_api import keyboard

from .jsonExtension import StructByAction


class Keyboard(keyboard.VkKeyboard):
    colors = {"blue": "primary", "white": "default",
              "red": "negative", "green": "positive"}

    def __init__(self, buttons=None, strategy="default", one_time=False, inline=False):
        if buttons is None:
            buttons = []
        self.strategy = strategy
        super().__init__(one_time=one_time, inline=inline)
        self.button_index = 0
        if isinstance(buttons, list) or isinstance(buttons, ListExtension):
            self.add_from_list(buttons)
        elif type(buttons) is dict:
            self.add_from_dict(buttons)
        elif type(buttons) is str:
            self.add_button(buttons, color="green")
        elif isinstance(buttons, StructByAction):
            self.add_from_list(buttons.dictionary)

    @classmethod
    def byKeyboard(cls, newKb):
        if isinstance(newKb, cls):
            return newKb
        else:
            return cls(newKb)

    def get_keyboard(self):
        self.cleanup_empty()
        return super().get_keyboard()

    @classmethod
    def get_empty_keyboard(cls):
        return cls()

    def add_button(self, *args, **kwargs):
        if getattr(Strategies, self.strategy)(self):
            self.add_line()
        args = list(args)
        if len(args) >= 2:
            color = args[1]
            del args[1]
        else:
            color = kwargs.get("color")
        if color is not None:
            if color in self.colors:
                color = self.colors[color]
        else:
            color = self.colors["white"]
        kwargs["color"] = color
        super().add_button(*args, **kwargs)
        self.button_index += 1
        return self

    def cleanup_empty(self):
        for i, line in enumerate(self.lines):
            if not line:
                del self.lines[i]

    def add_line(self):
        self.cleanup_empty()
        self.button_index = 0
        super().add_line()
        return self

    def add_from_list(self, arr):
        for j in arr:
            if j != "locationButton":
                self.add_button(j, color=self.get_random_color())
            else:
                self.add_location_button()
                self.add_line()

    def add_from_dict(self, d):
        if (strategy := d.get('strategy')) is not None:
            self.strategy = strategy
            del d['strategy']
        for k, v in d.items():
            if v == "line":
                self.add_line()
            else:
                self.add_button(k, v)

    def get_random_color(self):
        return random.choice(list(self.colors))

    def __repr__(self):
        return self.get_keyboard()

    # += operator
    def __iadd__(self, other):
        return self.__add__(other)

    def __add__(self, other):
        if isinstance(other, dict):
            self.add_from_dict(other)
        elif isinstance(other, list):
            self.add_from_list(other)
        else:
            self.add_button(other, color=self.get_random_color())
        return self


class Strategies(object):
    @staticmethod
    def default(keyboard: Keyboard):
        return keyboard.button_index % 4 == 0

    @staticmethod
    def insert_lines(_):
        return True

    @staticmethod
    def max_two_buttons(keyboard: Keyboard):
        return keyboard.button_index % 2 == 0



from typing import Iterable


class SetExtension(set):
    def __iadd__(self, other: Iterable):
        for element in other:
            self.add(element)
        return self

    def addAll(self, other):
        return self.__iadd__(other)



import inspect
from .dataClass import DataClass


# kotlin-like enums
class Enum(object):
    def __init__(self, customDataClass=None):
        inspectedArgs = inspect.getfullargspec(self.__init__)
        self.EnumValue = DataClass(
            *inspectedArgs.args[1:]) if customDataClass is None else customDataClass(*inspectedArgs.args[1:])



class DataClass(object):
    def __init__(self, *args, customClass=None):
        self.klass = DataClass if customClass is None else customClass
        self.args = args
        for i in args:
            self.__setattr__(i, None)

    def __call__(self, *args):
        newInstance = self.klass(*self.args)
        for j, i in enumerate(self.args):
            newInstance.__setattr__(i, args[j])
        return newInstance




class Event(object):
    callbacks = {}

    def __init__(self, name, func, param="") -> None:
        self.name = name
        self.param = param
        self.func = func
        callbacks = self.callbacks.get(name)
        if callbacks is None:
            Event.callbacks[name] = [self]
        else:
            Event.callbacks[name].append(self)
        super().__init__()

def on(name):
    def func_wrap(func):
        Event(name, func, "on")
    return func_wrap

def once(name):
    def func_wrap(func):
        Event(name, func, "once")
    return func_wrap


def emit(name, *args, **kwargs):
    callbacks = Event.callbacks.get(name)
    if callbacks is not None:
        for iterable, callback in enumerate(callbacks):
            callback.func(*args, **kwargs)
            if callback.param == "once":
                del Event.callbacks[name][iterable]



import threading
from . import events
from . import listExtension


class ThreadManager(object):
    thread_poll = listExtension.ListExtension()
    
    @classmethod
    def get_main_thread(cls):
        return cls.threadByName("Main")

    @classmethod
    def threadByName(cls, name):
        return cls.thread_poll.find(lambda item: item.name == name)

    def changeInterval(self, name, newInterval):
        thread = self.threadByName(name)
        thread.interval = newInterval

    @classmethod
    def create_task(cls, name, task, *args, **kwargs):
        thread = cls.threadByName(name)
        thread.create_task(task, *args, **kwargs)

    def __getitem__(self, key):
        return self.threadByName(key)


class Thread(threading.Thread):
    def __init__(self, *args, **kwargs) -> None:
        ThreadManager.thread_poll.append(self)
        self.tasks = []
        super().__init__(*args, **kwargs)

    def create_task(self, task, *args, **kwargs):
        self.tasks.append((task, args, kwargs))

    def check_tasks(self):
        while len(self.tasks) > 0:
            task = self.tasks.pop(0)
            task[0](*task[1], **task[2])


def requires_start(func):
    def call_wrap(*args, **kwargs):
        if main_thread is None or not getattr(main_thread, "started", False):
            @events.once("start")
            def _():
                func(*args, **kwargs)
        else:
            func(*args, **kwargs)

    return call_wrap


def threaded(*args, **kwargs):
    def func_wrap(func):
        def call_wrap(*a, **b):
            Thread(*args, **kwargs, target=func, args=a, kwargs=b).start()
        return call_wrap
    return func_wrap


class Every(Thread):
    def __init__(self, interval, *args, onExecCallback=None, callback=None, **kwargs):
        if callback is not None:
            self.callback = callback
        elif hasattr(self, "loop"):
            self.callback = self.loop
        else:
            raise Exception("Callback wasn't provided.")
        self.interval = interval
        self.stopped = False
        self.event = threading.Event()
        self.onExecCallback = onExecCallback
        self.args = args
        super().__init__(**kwargs)
        self.start()

    # override
    def run(self):
        self.callback(*self.args)
        while not self.event.wait(self.interval) and not self.stopped:
            if self.onExecCallback is not None:
                self.onExecCallback()
            self.check_tasks()
            self.callback(*self.args)


def every(interval, *myArgs, callback=None, **myKwargs):
    def func_wrap(func):
        return Every(interval, *myArgs, onExecCallback=callback, **myKwargs, callback=func)

    return func_wrap

main_thread = None

@events.once("start")
def update_main_thread():
    global main_thread
    main_thread = ThreadManager.threadByName("Main")



# -*- coding: utf-8 -*-
"""
:authors: SPRAVEDLIVO
:license: GNU AFFERO GENERAL PUBLIC LICENSE Version 3, see LICENSE file

:copyright: (c) 2022 SPRAVEDLIVO
"""
from .lp import BotLongPoll, UserLongPoll
from .user import User
from .keyboard import Keyboard
from .cmd import start_command, command, after_text_matcher, after_func

__author__ = 'SPRAVEDLIVO'
__version__ = '1.6'
__email__ = 'admin@spravedlivo.dev'



import json
import os
from typing import Callable


class StructByAction(object):
    def __init__(self, initDict, parent=None, parent_key=None, action: Callable = lambda _: _):
        self.parent = parent
        self.dictionary = initDict
        self.parent_key = parent_key
        self.action = action

    def __setitem__(self, key, value):
        if self.parent is None:
            self.dictionary.__setitem__(key, value)
            self.action(self.dictionary)
        else:
            self.dictionary[key] = value
            self.parent.__setitem__(self.parent_key, self.dictionary)

    def __getitem__(self, key):
        tmp_return = self.dictionary[key]
        if isinstance(tmp_return, dict) or isinstance(tmp_return, list):
            return StructByAction(tmp_return, parent=self, parent_key=key, action=self.action)
        else:
            return tmp_return

    def __iter__(self):
        return self.dictionary.__iter__()

    def get(self, key):
        return self.dictionary[key]

    def __str__(self):
        return self.dictionary.__str__()

    def __repr__(self):
        return f"StructByAction({self.dictionary})"

    def __delitem__(self, item):
        del self.dictionary[item]
        self.action(self.dictionary)

    def __contains__(self, item):
        return self.dictionary.__contains__(item)

    # list methods
    def __len__(self):
        return len(self.dictionary)

    def append(self, value):
        self.dictionary.append(value)
        self.action(self.dictionary)

    def __iadd__(self, keys):
        self.dictionary += keys
        self.action(self.dictionary)
        return self

    def insert(self, index, value):
        self.dictionary.insert(index, value)
        self.action(self.dictionary)
        return self

    def __bool__(self):
        return len(self.dictionary) > 0 if isinstance(self.dictionary, list) else len(self.dictionary.keys()) > 0


def save(file, obj, indent=None):
    with open(file, "w", encoding="utf-8") as f:
        json.dump(obj, f, indent=indent)


def load(file, indent=None):
    with open(file, encoding="utf-8") as f:
        return StructByAction(json.load(f), action=lambda d: save(file, d, indent))


def loadAdvanced(file, ident=None, content=None):
    if content is not None and not os.path.exists(file):
        with open(file, "w", encoding="utf-8") as f:
            content = json.dumps(content) if isinstance(content, dict) else content
            f.write(content)
    return load(file, ident)


def isCastToFloatAvailable(data):
    try:
        float(data)
        return True
    except ValueError:
        return False


def isDeserializable(data):
    try:
        if isCastToFloatAvailable(data) or not (data.startswith("{") or data.startswith("[")):
            return {}, False
        return json.loads(data), True
    except (ValueError, TypeError):
        return {}, False



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="vk_sdk",
    version="1.6",
    description="Wrapper around vk_api library",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/SPRAVEDLIVO/VK-SDK",
    author="SPRAVEDLIVO",
    author_email="spravedlivo@dimden.dev",
    license="AGPLv3",
    packages=["vk_sdk"],
    python_requires='>=3.7',
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
    ],   
    include_package_data=True,
    install_requires=["pytz", "vk_api"],
    entry_points={
        "console_scripts": [
            "realpython=reader.__main__:main",
        ]
    },
)


import os

from importlib import util
import importlib


class ImportTools(object):
    ignore = ["__pycache__"]
    imported = []
    modules = {}

    def __init__(self, paths=None):
        if paths is None:
            paths = ["packages"]
        for path in paths:
            if path in self.imported:
                continue
            if not os.path.exists(path):
                os.makedirs(path)
            for file in os.listdir(path):
                if path in self.ignore:
                    continue
                thisPath = os.path.join(path, file)
                if os.path.isdir(thisPath):
                    continue
                self.imp_by_path(thisPath)
                self.imported.append(path)

    def reload(self, module):
        for k, v in self.modules.items():
            if k == module:
                self.modules[k] = importlib.reload(v)

    def reload_all(self):
        for k, v in self.modules.items():
            self.modules[k] = importlib.reload(v)

    @classmethod
    def imp_by_path(cls, path):
        if not path.endswith(".py"):
            path += ".py"
        module_name = os.path.splitext(os.path.basename(path))[0]
        spec = util.spec_from_file_location(module_name, path)
        foo = util.module_from_spec(spec)
        spec.loader.exec_module(foo)
        cls.modules[path.replace(os.path.sep, "/")] = foo
        return foo


require = ImportTools.imp_by_path


import time
from datetime import datetime

import pytz


class Timestamp(object):
    def __init__(self, timestamp=None, sync_tz="Europe/Moscow"):
        self.sync_tz = sync_tz
        self.timestamp = timestamp or time.time()
        zrh = pytz.timezone(self.sync_tz)
        tztime = zrh.localize(datetime.fromtimestamp(self.timestamp))
        tzfloat = (tztime - datetime(1970, 1, 1,
                   tzinfo=pytz.utc)).total_seconds()
        self.diff = self.timestamp - tzfloat

    def get_time(self):
        return self.timestamp + self.diff

    def __float__(self):
        return self.get_time()

    def passed(self):
        return time.time() >= self.get_time()

    @classmethod
    def now(cls):
        return cls(time.time())

    def prettyprint(self):
        return datetime.fromtimestamp(self.get_time()).strftime("%Y/%m/%d %H:%M:%S")
